#include <v_lib.h>
#include "math.h"

#define		SPACE_TIME	0
#define		SPACE_FREQ	1

void
matZeroRows(matrix_p m,u32_t row_start,u32_t row_end)
{
	u32_t r, row_len;
	V_ASSERT(m);
	if(row_start > row_end || row_start > m->rows)
		return;
	if(row_end > m->rows)
		row_end = m->rows;
	row_len = m->cols * sizeof(vec_t);
	for(r = row_start;r < row_end;r++)
		memZero(m->cell[r],row_len);
}

/*
 * matRVMFTZeroRows(): same as matZeroRows(), but operates on
 * Fourier transform of real matrix.
 */
void
matRVMFTZeroRows(matrix_p rvmft,u32_t row_start,u32_t row_end)
{
	u32_t r, ft_len, row_bytes;
	V_ASSERT(rvmft);
	V_ASSERT(row_start <= row_end);
	/* Clamp the range to positive part of Fourier transform */
	ft_len = rvmft->rows >> 1;
	if(row_end > ft_len)
		row_end = ft_len;
	row_bytes = rvmft->cols * sizeof(vec_t);
	if(!row_start)
	{
		/* Clear zero frequency */
		memZero(rvmft->cell[0],row_bytes);
		row_start++;
	}
	/* Clear both positive and negative frequencies */
	for(r = row_start;r < row_end;r++)
	{
		memZero(rvmft->cell[r],row_bytes);
		memZero(rvmft->cell[rvmft->rows - r],row_bytes);
	}
	/* Clear highest frequency */
	if(row_end == ft_len)
		memZero(rvmft->cell[ft_len],row_bytes);
}

/*
 * fixme: this is weird
 */
void
matZeroCols(matrix_p m,u32_t col_start,u32_t col_end)
{
	V_ASSERT(m);
	
	if(m->rows == m->cols)
	{
		matTransposeIP(m);
		matZeroRows(m,col_start,col_end);
		matTransposeIP(m);
	}
	else
	{
		matrix_p transposed = matAlloc(m->cols,m->rows);
		matTranspose(transposed,m);
		matZeroRows(transposed,col_start,col_end);
		matTranspose(m,transposed);
		memFree(transposed);
	}
}

void
matRVMFTZeroCols(matrix_p rvmft,u32_t col_start,u32_t col_end)
{
	u32_t r, ft_len, row_bytes;
	V_ASSERT(rvmft);
	V_ASSERT(col_start <= col_end);
	/* Clamp the range to positive part of Fourier transform */
	ft_len = rvmft->cols >> 1;
	if(col_end >= ft_len)
		col_end = ft_len - 1;
	if(col_start == col_end)
	{
		for(r = 0;r < rvmft->rows;r++)
		{
			rvmft->cell[r][col_start << 1] = 0;
			rvmft->cell[r][(col_start << 1) + 1] = 0;
		}
	}
	else
	{
		row_bytes = (col_end - col_start + 1) * sizeof(vec_t) * 2;
		for(r = 0;r < rvmft->rows;r++)
			memZero(&rvmft->cell[r][col_start << 1],row_bytes);
	}
}

void
matBandReject(matrix_p m,u32_t lo_bound,u32_t hi_bound)
{
	matZeroCols(m,lo_bound,hi_bound);
	matZeroRows(m,lo_bound,hi_bound);
}

void
matBandPass(matrix_p m,u32_t lo_bound,u32_t hi_bound)
{
	if(lo_bound)
		matBandReject(m,0,lo_bound-1);
	if(hi_bound < m->rows-1)
		matZeroRows(m,hi_bound+1,m->rows-1);
	if(hi_bound < m->cols-1)
		matZeroCols(m,hi_bound+1,m->cols-1);
}

void
matOffset(matrix_p m,vec_t offset)
{
	u32_t r, c;
	V_ASSERT(m);
	for(r = 0;r < m->rows;r++)
		for(c = 0;c < m->cols;c++)
			m->cell[r][c] += offset;
}

void
vecRunningSum(vector_p dest,vector_p src)
{
	u32_t i;
    vec_p d = dest->elements, s = src->elements;
	V_ASSERT(src->size == dest->size);
    d[0] = s[0];
    for(i = 1;i < dest->size;i++)
    	d[i] = s[i] + d[i-1];
}

void
vecLinSpaced(vector_p dest,vec_t min,vec_t max)
{
	u32_t i, n_divs;
    vec_t delta;
    n_divs = dest->size;
    if(n_divs > 1)
    {
   		dest->elements[0] = min;
	    delta = (max - min) / (n_divs - 1);
	    for(i = 1;i < n_divs;i++)
   			dest->elements[i] = dest->elements[i - 1] + delta;
   	}
   	else
   		dest->elements[0] = (min + max) / 2;
}

void
matLinSpacedRows(matrix_p dest,vec_t min,vec_t max)
{
	u32_t i, n_divs, row_bytes;
    vec_t delta;
    n_divs = dest->cols;
	if(n_divs > 1)
	{
    	dest->cell[0][0] = min;
        delta = (max - min) / (n_divs - 1);
	    for(i = 1;i < dest->cols;i++)
    		dest->cell[0][i] = dest->cell[0][i-1] + delta;
    }
    else
    	dest->cell[0][0] = (min + max) / 2;
    row_bytes = n_divs * sizeof(vec_t);
    for(i = 1;i < dest->rows;i++)
    	memCopy(&(dest->cell[i][0]),&(dest->cell[0][0]),row_bytes);
}

void
matLinSpacedCols(matrix_p dest,vec_t min,vec_t max)
{
	u32_t i, j, n_divs;
    vec_t delta;
    n_divs = dest->rows;
    if(n_divs > 1)
    {
	    dest->cell[0][0] = min;
	    delta = (max - min) / (n_divs - 1);
    	for(i = 1;i < dest->rows;i++)
    		dest->cell[i][0] = dest->cell[i-1][0] + delta;
    }
    else
	    dest->cell[0][0] = (min + max) / 2;
    for(i = 1;i < dest->cols;i++)
    {
    	for(j = 0;j < dest->rows;j++)
			dest->cell[j][i] = dest->cell[j][0];
    }
}

void
matScaleAdd(matrix_p dest,matrix_p src,vec_t scale,vec_t offset)
{
	u32_t r, c;
	V_ASSERT(dest->rows == src->rows);
	V_ASSERT(dest->cols == src->cols);
	for(r = 0;r < dest->rows;r++)
		for(c = 0;c < dest->cols;c++)
			dest->cell[r][c] = offset + src->cell[r][c] * scale;
}

void
vecScaleAdd(vector_p dest,vector_p src,vec_t scale,vec_t offset)
{
	u32_t c;
	V_ASSERT(dest->size == src->size);
	for(c = 0;c < dest->size;c++)
		dest->elements[c] = offset + src->elements[c] * scale;
}

void
matScaleM(matrix_p dest,matrix_p src,matrix_p scale)
{
	u32_t r, c;
	V_ASSERT(dest->rows == src->rows);
	V_ASSERT(dest->cols == src->cols);
	V_ASSERT(dest->rows == scale->rows);
	V_ASSERT(dest->cols == scale->cols);
	for(r = 0;r < dest->rows;r++)
		for(c = 0;c < dest->cols;c++)
			dest->cell[r][c] = src->cell[r][c] * scale->cell[r][c];
}

void
matOffsetM(matrix_p dest,matrix_p src,matrix_p offset)
{
	u32_t r, c;
	V_ASSERT(dest->rows == src->rows);
	V_ASSERT(dest->cols == src->cols);
	V_ASSERT(dest->rows == offset->rows);
	V_ASSERT(dest->cols == offset->cols);
	for(r = 0;r < dest->rows;r++)
		for(c = 0;c < dest->cols;c++)
			dest->cell[r][c] = src->cell[r][c] + offset->cell[r][c];
}

/*
 * matHistogram()
 */
vector_p	histogram_space = NULL; /* fixme */

vboolean
matHistogram(vector_p dest,matrix_p src,vboolean normalize)
{
	u32_t rn, cn, rx, cx;
	u32_t sum, i, r, c, n_divs = dest->size;
    vec_t lower, upper, min, max, delta, *row;
    matMin(src,&rn,&cn);
    matMax(src,&rx,&cx);
    min = src->cell[rn][cn];
    max = src->cell[rx][cx];
    delta = (max - min) / n_divs;
    n_divs--;
    for(i = 0;i < n_divs;i++)
    {
    	sum = 0;
        lower = min + i * delta;
        upper = lower + delta;
        if(histogram_space)
        	histogram_space->elements[i] = (lower + upper) / 2;
        for(r = 0;r < src->rows;r++)
        {
        	row = src->cell[r];
        	for(c = 0;c < src->cols;c++)
            {
            	if(row[c] >= lower && row[c] < upper)
                	sum++;
        	}
		}
        dest->elements[i] = sum;
    }
    /* last bin */
	sum = 0;
    lower = min + i * delta;
	if(histogram_space)
		histogram_space->elements[i] = (2 * lower + delta) / 2;
    for(r = 0;r < src->rows;r++)
    {
    	row = src->cell[r];
    	for(c = 0;c < src->cols;c++)
        {
        	if(row[c] >= lower)
            	sum++;
    	}
	}
    dest->elements[i] = sum;
    n_divs++;
    if(normalize)
    {
    	vec_t vsum = dest->elements[0];
        for(i = 1;i < n_divs;i++)
        	vsum += dest->elements[i];
		if(vsum > 0)
        {
        	vsum = 1 / vsum;
            vecScale(dest,dest,vsum);
        }
        else
        	return false;
    }
    return true;
}

/*
 *
 */
vboolean
matAdjustContrast(matrix_p dest,matrix_p src,vec_t min,vec_t max)
{
	vec_t rscale;
    rscale = max - min;
    if(!IsZero(rscale))
    {
    	rscale = 1 / rscale;
    	matScaleAdd(dest,src,rscale,- min * rscale);
    	return true;
    }
   	conlPuts(LOG_WARNING,"cspaceAdjustContrast(): zero input range, can't transform\n");
    return false;
}

/*
 *
 */
vboolean
matMaximizeContrast(matrix_p dest,matrix_p src)
{
	u32_t r, c;
	vec_t min, max;
    matMin(src,&r,&c);
    min = src->cell[r][c];
    matMax(src,&r,&c);
    max = src->cell[r][c];
    return matAdjustContrast(dest,src,min,max);
}

/*
 *
 */
vboolean
matAdjustContrastV(vector_p dest,vector_p src,vec_t min,vec_t max)
{
	vec_t rscale;
    rscale = max - min;
    if(!IsZero(rscale))
    {
    	rscale = 1 / rscale;
    	vecScaleAdd(dest,src,rscale,- min * rscale);
    	return true;
    }
   	conlPuts(LOG_WARNING,"cspaceAdjustContrastV(): zero input range, can't transform");
    return false;
}

/*
 *
 */
vboolean
matMaximizeContrastV(vector_p dest,vector_p src)
{
	vec_t min, max;
    min = src->elements[vecMin(src)];
    max = src->elements[vecMax(src)];
    return matAdjustContrastV(dest,src,min,max);
}

/*
 *
 */
vboolean
matGammaCorrection(matrix_p dest,matrix_p src,vector_p gamma,
	vboolean unit_input)
{
	u32_t i, r, c, n_divs = gamma->size;
    vec_t min, max, t, delta, *rs, *rd;
    if(unit_input)
    {
    	min = 0;
        max = 1;
	}
    else
    {
		u32_t rn, cn, rx, cx;
    	matMin(src,&rn,&cn);
    	matMax(src,&rx,&cx);
    	min = src->cell[rn][cn];
    	max = src->cell[rx][cx];
	}
    delta = (max - min) / n_divs;
    if(IsZero(delta))
    	return false;
    for(r = 0;r < dest->rows;r++)
    {
    	rd = dest->cell[r];
    	rs = src->cell[r];
        for(c = 0;c < dest->cols;c++)
        {
        	t = (rs[c] - min) / delta; /* 0 <= t < n_divs */
            if(t < 0)
            	t = 0;
            i = (u32_t)t;
            if(i >= n_divs)
            	i = n_divs - 1;
			rd[c] = gamma->elements[i];
		}
	}
    return true;
}

